#!/usr/bin/env python

"""Test MTU between two IPv6 non-measurement machines."""

__author__ = 'Jeffrey Knockel'
__copyright__ = 'Copyright 2013 Jeffrey Knockel'
__license__ = 'GPL v2'

import sys

from icmp6 import receive_icmp
from intervention import intervention_at_time
from nslookup6 import lookup_or_die
from scanerror import ScanError
from sequence import ICMP6Sequence
from sniffing import start_sniffing
from toobig6 import too_big

INTERVENTION = 4
SAMPLES = 20
TIMEOUT = 1.0

def test_mtu(sniffer, addr1, addr2, mtu):
    """Return intervention analysis results for test of whether MTU from
    'addr1' to 'addr2' is no more than 'mtu'.

    This test forges echo requests of size 'mtu' from 'addr2' to addr1'.  This
    test then uses addr1's globally incrementing IPv6 id as a side-channel, as
    this will only increment when addr1 has to fragment (and thus include an
    IPv6 fragmentation extension header) the echo replies to addr2.

    For interpreting the p-value in the result, the caller is recommended to
    choose some significance level 'alpha', e.g., 0.01.  If the resulting
    p-value is less than 'alpha', then the measured MTU may be interpreted as
    no more than 'mtu', whereas if the resulting p-value is greater than 1 -
    'alpha', then the measured MTU may be interpreted as greater than 'mtu'.
    If the p-value falls outside these ranges, then the results are not
    statistically significant.
    """
    last_ip_id = None
    xs = []
    seq = ICMP6Sequence(addr1, iface=sniffer.iface)
    seq_spoofed = ICMP6Sequence(addr1, addr2)
    timeout = sniffer.timeout()
    for i in xrange(SAMPLES + 1):
        timeout.add_time(TIMEOUT)
        spoofed_mtu = mtu if i > SAMPLES // 2 else 0
        for _ in xrange(INTERVENTION):
            spoofed = seq_spoofed.next_packet_testing_mtu(spoofed_mtu)
            sniffer.send(spoofed, True)
        request = seq.next_packet_testing_mtu(1508)
        sniffer.send(request, True)
        sniffer.heartbeat()
        p = receive_icmp(sniffer, request, timeout, True)
        ip_id = None
        if p is not None:
            frag = p.extension_hdrs[44]
            if frag is not None:
                ip_id = frag.id
            else:
                sniffer.send(too_big(p, 1500, request.src, request.dst))
        if i > 0:
            diff = None if ip_id is None or last_ip_id is None else ip_id - last_ip_id - 1 & 0xffffffff
            sys.stderr.write('%s%s: %s\n' % (ip_id, '*' * (i > SAMPLES // 2), diff))
            xs.append(diff)
        last_ip_id = ip_id
        timeout.sleep()
    if all(x is None for x in xs):
        raise ScanError('Cannot get \'from\' host\'s IP id differences--does it respond to large ICMP echo requests?')
    try:
        return intervention_at_time(xs, [SAMPLES // 2], [INTERVENTION / 2.0], 0.05, True)
    except Exception:
        raise ScanError('The \'from\' host\'s traffic too noisy or lossy--try another host?')

def usage(code=1):
    sys.stderr.write('Usage: %s HOST1 HOST2 MTU\n' % sys.argv[0])
    exit(code)

def main():
    try:
        if len(sys.argv) != 4:
            usage()
        host1 = sys.argv[1]
        host2 = sys.argv[2]
        try:
            mtu = int(sys.argv[3])
        except ValueError:
            usage()
        addr1 = lookup_or_die(host1)[0]
        addr2 = lookup_or_die(host2)[0]
        ir = start_sniffing(test_mtu, True, args=(addr1, addr2, mtu))
        print ir
    except ScanError as e:
        sys.stderr.write('Error performing scan: %s\n' % e)
    except KeyboardInterrupt:
        sys.stderr.write('Received SIGINT\n')

if __name__ == '__main__':
    main()
