#!/usr/bin/env python

"""Measure SYN/ACK retransmission patterns."""

__author__ = 'Jeffrey Knockel'
__copyright__ = 'Copyright 2013 Jeffrey Knockel'
__license__ = 'GPL v2'

import sys
import time

from dpkt.tcp import TH_SYN, TH_ACK

from nslookup4 import lookup_or_die
from sequence import TCPSequence
from sniffing import start_sniffing, start_session
from tcp4 import receive_tcps

@start_session
def synack_retransmit_times(sniffer, addr, spoofed_addr, port, window=20.0, repeat=1, seconds=10, verbose=False):
    """Returns the SYN/ACK retransmission times from 'addr' by sending
    spoofed SYN's from 'spoofed_addr' on port 'port'.

    Choose 'spoofed_addr' as an unused address on your subnet that is not bound
    to your (or anyone's) interface so that your kernel will not respond to the
    incoming SYN/ACK's with RST's.   This address will be ARP-spoofed so that
    responses will still be sniffed.
    SYN/ACK retransmissions will be listened to for up to 'window' seconds
    (default 20).  SYN's will be transmitted every second for 'seconds' seconds
    (default 10) sending 'repeat'-many SYN's (default 1) per second.

    This function also returns a dict of statistics where key 'histo' is a list
    where index i is the number of SYN's that received i SYN/ACK's, key
    'avg_trans' is the average number of SYN/ACK's received per second, and key
    'syn_frac', the fraction of SYN's that received any SYN/ACK's.
    """
    sniffer.spoof_addr_add(spoofed_addr)
    seq = TCPSequence(addr, spoofed_addr)
    timeout = sniffer.timeout()
    requests = []
    sport_to_times = {}
    since = time.time() if verbose else None
    def check_packet(p):
        start, times = sport_to_times[p.data.dport]
        delay = time.time() - start
        if delay < window:
            if p.data.flags & TH_SYN and p.data.flags & TH_ACK:
                times.append(delay)
            if verbose:
                now = time.time() - since
                sys.stderr.write('%f: %s on port %d with id %d after %f seconds\n' % (now, p.data.flags, p.data.dport, p.id, delay))
    for _ in xrange(seconds):
        timeout.add_time(1.0)
        t = time.time()
        for _ in xrange(repeat):
            request = seq.next_packet(dport=port)
            sniffer.send(request)
            requests.append(request)
            sport_to_times[request.data.sport] = t, []
        for p in receive_tcps(sniffer, requests, timeout):
            check_packet(p)
    timeout.add_time(window)
    for p in receive_tcps(sniffer, requests, timeout):
        check_packet(p)
    for request in requests:
        seq.rst(request)
        sniffer.send(request)
    all_times = [sorted(times) for _, times in sport_to_times.itervalues()]
    max_len = max(len(times) for times in all_times)
    retransmit_times = [times for times in all_times if len(times) == max_len]
    n = len(retransmit_times)
    averaged = [sum(times) / n for times in zip(*retransmit_times)]
    histo = [0] * (max_len + 1)
    for times in all_times:
        histo[len(times)] += 1
    syns = repeat * seconds
    avg_trans = float(sum(len(times) for times in all_times)) / syns
    syn_frac = float(syns - histo[0]) / syns
    return averaged, {
        'histo': histo,
        'avg_trans': avg_trans * repeat,
        'syn_frac': syn_frac,
    }

def usage(code=1):
    sys.stderr.write('Usage: %s HOST PORT SPOOFED_HOST COUNT\n' % sys.argv[0])
    exit(code)

def main():
    try:
        if len(sys.argv) != 5:
            usage()
        host = sys.argv[1]
        try:
            port = int(sys.argv[2])
            repeat = int(sys.argv[4])
        except ValueError:
            usage()
        spoofed_host = sys.argv[3]
        addr = lookup_or_die(host)[0]
        spoofed_addr = lookup_or_die(spoofed_host)[0]
        retransmit_times, stats = start_sniffing(synack_retransmit_times, False, args=(addr, spoofed_addr, port), kwargs=dict(repeat=repeat, verbose=True))
        print 'Average SYNACK\'s transmitted per SYN: %f' % stats['avg_trans']
        print 'Fraction of SYN\'s transmitting any SYNACK: %f' % stats['syn_frac']
        print 'Histogram:'
        for i, n in enumerate(stats['histo']):
            print '%d: %d' % (i, n)
        print 'Transmission times:'
        print retransmit_times
    except KeyboardInterrupt:
        sys.stderr.write('Received SIGINT\n')

if __name__ == '__main__':
    main()
