#!/usr/bin/env python

"""Test the rate at which hosts will respond to TCP packets."""

__author__ = 'Jeffrey Knockel'
__copyright__ = 'Copyright 2013 Jeffrey Knockel'
__license__ = 'GPL v2'

import sys

from dpkt.tcp import TH_SYN, TH_ACK, TCPFlags

from nslookup4 import lookup_or_die
from sequence import TCPSequence
from sniffing import start_sniffing
from tcp4 import receive_tcps

def tcp_rate_test(sniffer, addr, repeat, seconds=10, sport=None, dport=None, flags=TCPFlags(TH_SYN), verbose=False):
    """Send 'repeat'-many TCP packets to 'addr' every second for 'seconds'
    seconds (default 10) and return a dict of statistics of the number of
    responses.

    In the returned dict, 'min' will be the minimum number of responses over
    all seconds, 'max' will be the maximum number, and 'avg' will be the
    average number.
    If 'sport', 'dport', or 'flags' are specified, these will be the source
    port, destination port, or flags used on the sent TCP packets.
    If 'verbose' is specified as True, the number of responses will be printed
    to stderr every second.
    """
    seq = TCPSequence(addr, iface=sniffer.iface)
    timeout = sniffer.timeout()
    min_count = float('inf')
    max_count = 0
    total_count = 0
    for _ in xrange(seconds):
        timeout.add_time(1.0)
        sports = set()
        requests = []
        for _ in xrange(repeat):
            request = seq.next_packet(sport=sport, dport=dport, flags=flags)
            sports.add(request.data.sport)
            requests.append(request)
            sniffer.send(request)
        for p in receive_tcps(sniffer, requests, timeout):
            sports.discard(p.data.dport)
        count = repeat - len(sports)
        if verbose:
            sys.stderr.write('%d/%d\n' % (count, repeat))
        min_count = min(min_count, count)
        max_count = max(max_count, count)
        total_count += count
    return {'min': min_count, 'max': max_count, 'avg': float(total_count) / seconds}

def usage(code=1):
    sys.stderr.write('Usage: %s HOST PORT COUNT\n' % sys.argv[0])
    exit(code)

def main():
    try:
        if len(sys.argv) != 4:
            usage()
        host = sys.argv[1]
        try:
            port = int(sys.argv[2])
            repeat = int(sys.argv[3])
        except ValueError:
            usage()
        addr = lookup_or_die(host)[0]
        ret = start_sniffing(tcp_rate_test, False, args=(addr, repeat), kwargs=dict(sport=port, flags=TCPFlags(TH_SYN | TH_ACK), verbose=True))
        print 'Min: %d' % ret['min']
        print 'Max: %d' % ret['max']
        print 'Avg: %.3g' % ret['avg']
    except KeyboardInterrupt:
        sys.stderr.write('Received SIGINT\n')

if __name__ == '__main__':
    main()
