#!/usr/bin/env python

__author__ = 'Jeffrey Knockel'
__copyright__ = 'Copyright 2013 Jeffrey Knockel'
__license__ = 'GPL v2'

import sys
import time

from dpkt.tcp import TH_SYN, TH_ACK, TCPFlags

from nslookup4 import lookup_or_die
from sequence import TCPSequence
from sniffing import start_sniffing
from tcp4 import receive_tcps

def tcp_detect_ip_id_pattern(sniffer, addrs, spoofed_addr, sport=None, dport=None, flags=TCPFlags(TH_SYN), threshold=5000, seconds=10, verbose=False):
    """Return the TCP reply IP id pattern for 'addrs' using 'spoofed_addr' to
    spoof TCP packets from in order to determine if the IP id pattern is
    incrementing per-host.

    Returns a dict mapping addresses to result dicts.  Each result dict has a
    key 'type' containing the name of the pattern, 'min_time' containing the
    earliest measured ping time, 'avg_time' containing the average ping time,
    and 'avg_diff' containing the average IP id difference between replies.
    The returned pattern will be one of
    'echoed' - if the IP id's are the same as in the request
    'per-host - if the IP id's are incremented independently for each host
    'same' - if the IP id's are the same (e.g., zero)
    'globally' - if the IP id's are chosen from a globally incrementing counter
    'random' - if the IP id's are chosen randomly
    'inconclusive' - if the host responded with too few replies
    'none' - if the host did not respond with any replies
    The average IP id difference between replies must be over 'threshold'
    (default 5000) for the IP id to be considered random.  A request to each
    host is sent every second for 'seconds' seconds (default 10).
    If 'verbose' is specified as True, the received IP id will be printed to
    stderr for every second ('*' if no id is received).
    """
    threshold += 1 # to account for our spoofed packet
    results = dict((addr, {
        'seq': TCPSequence(addr, iface=sniffer.iface),
        'seq_spoofed': TCPSequence(addr, spoofed_addr),
        'ip_id1': None,
        'n': 0,
        'total_diff': 0,
        'are_incrementing_by_one': True,
        'are_same': True,
        'are_echoed': True,
        'min_time': float('inf'),
        'total_time': 0.0,
    }) for addr in addrs)
    timeout = sniffer.timeout()
    for _ in xrange(seconds):
        timeout.add_time(1.0)
        requests = []
        start = time.time()
        for addr, r in results.iteritems():
            request = r['seq'].next_packet(sport=sport, dport=dport, flags=flags)
            requests.append(request)
            sniffer.send(request)
            spoofed = r['seq_spoofed'].next_packet(sport=sport, dport=dport, flags=flags)
            sniffer.send(spoofed)
            r['ip_id2'] = None
        for p in receive_tcps(sniffer, requests, timeout):
            ping_time = time.time() - start
            if p.src in results:
                r = results[p.src]
                if r['ip_id2'] is None:
                    r['ip_id2'] = p.id
                    if r['ip_id1'] is not None:
                        r['min_time'] = min(r['min_time'], ping_time)
                        r['total_time'] += ping_time
                        diff = (r['ip_id2'] - r['ip_id1']) & 0xffff
                        r['total_diff'] += diff
                        r['are_same'] = r['are_same'] and diff == 0
                        r['are_incrementing_by_one'] = r['are_incrementing_by_one'] and diff == 1
                        r['are_echoed'] = r['are_echoed'] and request.id == r['ip_id2']
                        r['n'] += 1
                    if verbose:
                        if len(addrs) > 1:
                            sys.stderr.write('%s: %d\n' % (p.src.ascii, p.id))
                        else:
                            sys.stderr.write('%d\n' % p.id)
        if verbose:
            for addr, r in results.iteritems():
                if r['ip_id2'] is None:
                    if len(addrs) > 1:
                        sys.stderr.write('%s: *\n' % addr.ascii)
                    else:
                        sys.stderr.write('*\n')
        for _, r in results.iteritems():
            r['ip_id1'] = r['ip_id2']
    ret = {}
    for addr, r in results.iteritems():
        avg_diff = r['total_diff'] / float(r['n']) if r['n'] > 0 else float('nan')
        if r['n'] >= 5:
            if r['are_echoed']:
                type = 'echoed'
            if r['are_incrementing_by_one']:
                type = 'per-host'
            elif r['are_same']:
                type = 'same'
            elif avg_diff <= threshold:
                type = 'globally'
            else:
                type = 'random'
        elif r['n'] > 0:
            type = 'inconclusive'
        else:
            type = 'none'
        min_time = r['min_time']
        avg_time = r['total_time'] / r['n'] if r['n'] > 0 else float('nan')
        ret[addr] = {
            'type': type,
            'min_time': min_time,
            'avg_time': avg_time,
            'avg_diff': avg_diff,
        }
    return ret

def tests(sniffer, addr, spoofed_addr, port):
    print 'Testing %s...' % addr.ascii
    flags = TCPFlags(TH_SYN | TH_ACK)
    ip_id_pattern = tcp_detect_ip_id_pattern(sniffer, [addr], spoofed_addr, sport=port, flags=flags, verbose=True)[addr]
    min_time = ip_id_pattern['min_time']
    avg_time = ip_id_pattern['avg_time']
    print 'TCP %s port %d IP id pattern type: %s' % (flags, port, ip_id_pattern['type'])
    print 'TCP %s port %d IP id minimum latency: %.4g ms' % (flags, port, round(min_time * 1000))
    print 'TCP %s port %d IP id average latency: %.4g ms' % (flags, port, round(avg_time * 1000))

def usage(code=1):
    sys.stderr.write('Usage: %s HOST PORT SPOOFED_HOST\n' % sys.argv[0])
    exit(code)

def main():
    try:
        if len(sys.argv) != 4:
            usage()
        host = sys.argv[1]
        try:
            port = int(sys.argv[2])
        except ValueError:
            usage()
        spoofed_host = sys.argv[3]
        addr = lookup_or_die(host)[0]
        spoofed_addr = lookup_or_die(spoofed_host)[0]
        start_sniffing(tests, False, args=(addr, spoofed_addr, port))
    except KeyboardInterrupt:
        sys.stderr.write('Received SIGINT\n')

if __name__ == '__main__':
    main()
